//
// Created by philipp on 30.11.17.
//

#ifndef PROTOCOL_LEADERVEHICLE_H
#define PROTOCOL_LEADERVEHICLE_H


#include <mutex>
#include <unordered_map>
#include <bits/unordered_map.h>
#include "Vehicle.h"
#include "PlatoonService.h"
#include "Utils.h"
#include "NetworkingLib/DatagramSender.h"
#include "NetworkingLib/DatagramReceiver.h"

namespace platoonProtocol
{

class LeaderVehicle : public Vehicle
{
private:
    struct PrivateTag
    {
    };

public:
    static constexpr networking::time::Duration HEARTBEAT_TIMEOUT = std::chrono::milliseconds(1000);
    static constexpr networking::time::Duration BROADCAST_INTERVAL = std::chrono::milliseconds(50);

    using Ptr = std::shared_ptr<LeaderVehicle>;

    static Ptr create(networking::Networking & net, const NetworkInfo & info);

    // This should not be used outside.
    LeaderVehicle(PrivateTag, networking::Networking & net, const NetworkInfo & info);

    void stop() override;

    std::size_t getNumFollowers();

    void setPlatoonConfig(const PlatoonConfig & platoonConfig);

    void setPlatoonSpeed(PlatoonSpeed platoonSpeed);

    void setInnerPlatoonDistance(InnerPlatoonDistance innerPlatoonDistance);

protected:
    void doCreatePlatoon() override;

    void doRunPlatoon() override;

    void doLeavePlatoon() override;

    void receiveMessage(const PlatoonMessage & message) override;

    Ptr shared_from_this();

private:
    struct Follower
    {
        VehicleId vehicleId;
        networking::time::Timer::Ptr heartbeatTimer;
    };

    std::unordered_map<VehicleId, Follower> followers{2};
    std::mutex followersMutex;
    networking::time::Timer::Ptr broadcastTimer;

    void receivePlatoonCreateRequest(const PlatoonMessage & request);

    bool findFollower(VehicleId id) const;

    void addFollower(VehicleId follower);

    void removeFollower(Follower & follower);

    void startHeartbeatTimer(Follower & follower);

    void stopHeartbeatTimer(Follower & follower);

    void receiveHeartbeat(const PlatoonMessage & heartbeat);

    void receiveLeavePlatoonMessage(const PlatoonMessage & message);

    void startBroadcasting();

    void stopBroadcasting();

    void sendBroadcastMessage();

    void removeAllFollowers();

    bool checkFollowerMessage(const PlatoonMessage & message) const;
};

}


#endif //PROTOCOL_LEADERVEHICLE_H
